Public Delegate Function MyFxDelegate(ByVal nNumVars As Integer, ByRef fX() As Double, ByRef fParam() As Double) As Double
Public Delegate Function SayFxDelegate() As String

Public Class CMarquardt1
  Dim m_MyFx As MyFxDelegate

  Protected Function MyFxEx(ByVal nNumVars As Integer, _
              ByRef fX() As Double, ByRef fParam() As Double, _
              ByRef fDeltaX() As Double, ByVal fLambda As Double) As Double
    Dim I As Integer
    Dim fXX(nNumVars) As Double

    For I = 0 To nNumVars - 1
      fXX(I) = fX(I) + fLambda * fDeltaX(I)
    Next I

    MyFxEx = m_MyFx(nNumVars, fXX, fParam)
  End Function

  Protected Function FirstDeriv(ByVal nNumVars As Integer, _
               ByRef fX() As Double, ByRef fParam() As Double, _
               ByVal nIdxI As Integer) As Double
    Dim fXt, h, Fp, Fm As Double

    fXt = fX(nIdxI)
    h = 0.01 * (1 + Math.Abs(fXt))
    fX(nIdxI) = fXt + h
    Fp = m_MyFx(nNumVars, fX, fParam)
    fX(nIdxI) = fXt - h
    Fm = m_MyFx(nNumVars, fX, fParam)
    fX(nIdxI) = fXt
    FirstDeriv = (Fp - Fm) / 2 / h
  End Function

  Protected Function SecondDeriv(ByVal nNumVars As Integer, _
                 ByRef fX() As Double, ByRef fParam() As Double, _
                 ByVal nIdxI As Integer, ByVal nIdxJ As Integer) As Double
    Dim fXt, fYt, fHX, fHY, F0, Fp, Fm As Double
    Dim Fpp, Fmm, Fpm, Fmp, fResult As Double

    ' calculate second derivative?
    If nIdxI = nIdxJ Then
      F0 = m_MyFx(nNumVars, fX, fParam)
      fXt = fX(nIdxI)
      fHX = 0.01 * (1 + Math.Abs(fXt))
      fX(nIdxI) = fXt + fHX
      Fp = m_MyFx(nNumVars, fX, fParam)
      fX(nIdxI) = fXt - fHX
      Fm = m_MyFx(nNumVars, fX, fParam)
      fX(nIdxI) = fXt
      fResult = (Fp - 2 * F0 + Fm) / fHX ^ 2
    Else
      fXt = fX(nIdxI)
      fYt = fX(nIdxJ)
      fHX = 0.01 * (1 + Math.Abs(fXt))
      fHY = 0.01 * (1 + Math.Abs(fYt))
      ' calculate Fpp
      fX(nIdxI) = fXt + fHX
      fX(nIdxJ) = fYt + fHY
      Fpp = m_MyFx(nNumVars, fX, fParam)
      ' calculate Fmm
      fX(nIdxI) = fXt - fHX
      fX(nIdxJ) = fYt - fHY
      Fmm = m_MyFx(nNumVars, fX, fParam)
      ' calculate Fpm
      fX(nIdxI) = fXt + fHX
      fX(nIdxJ) = fYt - fHY
      Fpm = m_MyFx(nNumVars, fX, fParam)
      ' calculate Fmp
      fX(nIdxI) = fXt - fHX
      fX(nIdxJ) = fYt + fHY
      Fmp = m_MyFx(nNumVars, fX, fParam)
      fX(nIdxI) = fXt
      fX(nIdxJ) = fYt
      fResult = (Fpp - Fmp - Fpm + Fmm) / (4 * fHX * fHY)
    End If

    Return fResult
  End Function


  Protected Sub GetFirstDerives(ByVal nNumVars As Integer, _
                  ByRef fX() As Double, ByRef fParam() As Double, _
                  ByRef fFirstDerivX() As Double)

    Dim I As Integer

    For I = 0 To nNumVars - 1
      fFirstDerivX(I) = FirstDeriv(nNumVars, fX, fParam, I)
    Next I
  End Sub

  Protected Sub GetSecondDerives(ByVal nNumVars As Integer, _
                  ByRef fX() As Double, ByRef fParam() As Double, _
                  ByRef fSecondDerivX(,) As Double)

    Dim I, J As Integer

    For I = 0 To nNumVars - 1
      For J = 0 To nNumVars - 1
        fSecondDerivX(I, J) = SecondDeriv(nNumVars, fX, fParam, I, J)
      Next J
    Next I
  End Sub

  Protected Function LinSearch_DirectSearch(ByVal nNumVars As Integer, ByRef fX() As Double, ByRef fParam() As Double, _
            ByRef fLambda As Double, ByRef fDeltaX() As Double, ByVal fInitStep As Double, ByVal fMinStep As Double) As Boolean
    Dim F1, F2 As Double

    F1 = MyFxEx(nNumVars, fX, fParam, fDeltaX, fLambda)

    Do
      F2 = MyFxEx(nNumVars, fX, fParam, fDeltaX, fLambda + fInitStep)
      If F2 < F1 Then
        F1 = F2
        fLambda += fInitStep
      Else
        F2 = MyFxEx(nNumVars, fX, fParam, fDeltaX, fLambda - fInitStep)
        If F2 < F1 Then
          F1 = F2
          fLambda -= fInitStep
        Else
          ' reduce search step size
          fInitStep /= 10
        End If
      End If
    Loop Until fInitStep < fMinStep

    Return True

  End Function

  Public Function CalcOptim(ByVal nNumVars As Integer, ByRef fX() As Double, ByRef fParam() As Double, _
                  ByRef fToler() As Double, ByVal fEpsFx As Double, ByVal nMaxIter As Integer, _
                  ByRef fAlpha As Double, ByVal C1 As Double, ByVal C2 As Double, _
                  ByRef nIter As Integer, ByRef sErrorMsg As String, _
                  ByVal MyFx As MyFxDelegate) As Double

    Const ALPHA_MAX = 1.0E+100

    Dim I As Integer
    Dim F1, F2, fNorm As Double
    Dim nInnerLoopIter As Integer
    Dim bStop As Boolean
    Dim fX2(nNumVars) As Double
    Dim g(nNumVars) As Double
    Dim G2(nNumVars) As Double
    Dim DeltaX(nNumVars) As Double
    Dim J(nNumVars, nNumVars) As Double
    Dim J2(nNumVars, nNumVars) As Double
    Dim fIdenMat(nNumVars, nNumVars) As Double
    Dim fIdenMat2(nNumVars, nNumVars) As Double
    Dim nIndex(nNumVars) As Integer

    On Error GoTo HandleErr

    m_MyFx = MyFx

    ' validate iterations parameters
    If C1 <= 0 Or C1 >= 1 Then C1 = 0.1
    If C2 <= 1 Then C2 = 10

    F2 = MyFx(nNumVars, fX, fParam)

    nIter = 1
    Do
      nIter += 1
      If nIter > nMaxIter Then
        sErrorMsg = "Reached maximum iterations limit"
        Exit Do
      End If

      ' copy last Fx value
      F1 = F2

      GetFirstDerives(nNumVars, fX, fParam, g)
      MatrixLibVb.DuplicateVector(g, G2)

      ' test if gradient is shallow enough
      fNorm = 0
      For I = 0 To nNumVars - 1
        fNorm += g(I) ^ 2
      Next I
      fNorm = Math.Sqrt(fNorm)
      If fNorm < fEpsFx Then Exit Do

      MatrixLibVb.DuplicateVector(fX, fX2) ' copy fX
      ' get matrix J
      GetSecondDerives(nNumVars, fX, fParam, J)
      ' create identitty matrix
      MatrixLibVb.MatIdentity(fIdenMat)
      ' initialize inner-loop counter
      nInnerLoopIter = 0

      Do
        nInnerLoopIter += 1
        If nInnerLoopIter > nMaxIter Then Exit Function

        MatrixLibVb.DuplicateVector(G2, g)
        MatrixLibVb.MatMultVal(fIdenMat, fAlpha, fIdenMat2)
        MatrixLibVb.MatAddMat(J, fIdenMat2, J2)
        MatrixLibVb.MV_LUDecomp(J2, nIndex, nNumVars)
        MatrixLibVb.MV_LUBackSubst(J2, nIndex, nNumVars, g)
        For I = 0 To nNumVars - 1
          DeltaX(I) = g(I)
          fX(I) -= DeltaX(I)
        Next I

        F2 = MyFx(nNumVars, fX, fParam)

        If F2 < F1 Then
          fAlpha = C1 * fAlpha
          bStop = True
        Else
          fAlpha = C2 * fAlpha
          MatrixLibVb.DuplicateVector(fX2, fX) ' restore array fX
          bStop = False
          If fAlpha >= ALPHA_MAX Then Return F2
        End If
      Loop Until bStop

    Loop

ExitProc:
    Return F2

HandleErr:
    sErrorMsg = "Calculation error: " & Err.Description
    Resume ExitProc
  End Function

End Class
